#include <cstdlib>
#include <config.h>
#include <ChiliFileDecoder.h>
#include <ChiliSession.h>
#include <ChiliUISession.h>
#include <FL/fl_ask.H>
#include <ChiliLog.h>
#include <stdio.h>
#include <string.h>
#include <numtostr.h>
#include <ctype.h>
#include <sndfile.h>

bool FileExists(const string filename) {
	FILE * ptrF;
	ptrF = fopen(filename.c_str(), "r");
	if (ptrF!=0) {
		fclose(ptrF);
		return true;
	}
	return false;
}
string ChiliFileDecoder::generate_filename(string tempdir) {
	srand((unsigned) time(NULL));
	int j=(int) ((float)99999999*rand()/(RAND_MAX+1.0));
	string mynum="00000000"+numtostr(j);
	string strFilename=tempdir+mynum.substr(mynum.length()-8,8)+".raw";
	while (FileExists(strFilename)) {
		int j=(int) ((float)99999999*rand()/(RAND_MAX+1.0));
		mynum="00000000"+numtostr(j);
		strFilename=tempdir+mynum.substr(mynum.length()-8,8)+".raw";
	}
	return strFilename;
}
void ChiliFileDecoder::decode_file (SNDFILE *infile)
{      
	/* Given an inputfile, this will decode the file to the
	 * temporary directory using a random filename.
	 */
	log("ChiliFileDecoder::decode_file(infile,channels=%s)",numtostr(number_of_channels).c_str());
	string tempdir=session->GetTempDir();
	ChiliUISession * uisession=session->GetUISession(); //see if we have access to an UI
	
	string rawfilename=generate_filename(tempdir);
	tmpfilename=rawfilename;
	string shortrawname=rawfilename;	// will contain just filename without path

	shortrawname=shortrawname.substr(tempdir.length(),rawfilename.length()-tempdir.length());
	filekey=shortrawname;
	log("filekey=%s",shortrawname.c_str());
	int len=SNDFILE_BUFFER_LEN/number_of_channels; // max number of multichannel samples we can read
        long    readcount ;
        readcount = len;

	log("working filename=%s",rawfilename);
	log("source filename=%s",filename);
	
	/* In user prefs, store which raw working filename is 
	 * related connected with which source file.
	 * e.g. 
	 * 2231154.raw:/home/mrjb/bla/bla/we/rt/w/y/axxx.wav
	 * This allows us to load files from read-only and network
	 * media and edit them locally without losing their 
	 * relation with the source file. This will also allow us
	 * to recover the working file after a crash (as we will
	 * clean up the raw file when closing the project).
	 */
	
	int outfilemajor=SF_FORMAT_RAW;
	int outfileminor=SF_FORMAT_FLOAT;
	sfinfoout.format=outfilemajor|outfileminor|SF_ENDIAN_LITTLE;
	sfinfoout.channels=number_of_channels;
	sfinfoout.sections=1;
	sfinfoout.samplerate=44100;

	FILE * outfileptr;
	
	if (!(outfileptr=fopen(rawfilename.c_str(),"w"))) {
		log("Not able to open temporary file %s",rawfilename);
		puts (sf_strerror (NULL)) ;
		throw ERROR_CANNOT_OPEN_OUTPUTFILE;
	}
	int fd=fileno(outfileptr);	
//	if (!(outfile=sf_open(rawfilename.c_str(),SFM_WRITE,&sfinfoout))) {
	if (!(outfile=sf_open_fd(fd,SFM_WRITE,&sfinfoout,true))) {
		log("Not able to open temporary file %s",rawfilename);
		puts (sf_strerror (NULL)) ;
		throw ERROR_CANNOT_OPEN_OUTPUTFILE;
	}

	float progress=0;
	if (uisession!=NULL) {
		uisession->set_progress_pct(progress); // send 0% to init progress bar
	}
	int filelen=(int)sf_seek(infile,0,SEEK_END); // in multichannel samples 
	log ("FileLen is %s multichannel samples",numtostr(filelen).c_str());

	sf_seek(infile,0,SEEK_SET); // back to start

	int readblocks=0;
	int totblocks=1+(int)floor(  (filelen*number_of_channels)/len  );

        while (readcount > 0)   {
		readblocks++;
                readcount = sf_read_float (infile, data, len);
		if (uisession!=NULL) {
			uisession->set_progress_pct(progress);
			if (progress<100) { 
				if (totblocks==0) {
					progress=100;
				} else {
					progress=(100*readblocks/totblocks);
				}
			}
		}
		/* readcount=number of frames read
		 * This normally produces channel-interlaced data.
		 * We actually want to work with interlaced data
		 * for multichannel performance reasons, so no need
		 * to deinterlace.  However this is an ideal place
		 * for calculating peak data.
		 */
		
		if (readcount>0) {
	                sf_write_float (outfile, data, readcount); 
			fsync(fd);
		}
        };
	if (uisession!=NULL) {
		uisession->set_progress_pct(100); // terminate progress bar with 100% reading	
	}
	sf_close(outfile);
	log("ChiliFileDecoder::decode file is done.");
       	return;
} /* copy_data */

void ChiliFileDecoder::initvars() {
	number_of_channels=0;
	this->session=NULL;
	data=NULL;
	tmpdata=NULL;
	data=(float *)malloc(SNDFILE_BUFFER_LEN*sizeof(float));
	tmpdata=(float *)malloc(SNDFILE_BUFFER_LEN*sizeof(float));
	
	filename=*new string("");
	tmpfilename=*new string("");
	filekey=*new string("");
	outfile=NULL;	// SNDFILE * outfile
}

ChiliFileDecoder::ChiliFileDecoder(ChiliSession * p_session,string p_filename,const int samplerate,const int channels) {
	log("ChiliFileDecoder::ChiliFileDecoder(%s,samplerate,channels)",p_filename);
	initvars();
	if (p_session!=NULL) {
		session=p_session;		
	}
	if (p_filename=="") {
		/* When no filename is given, we may be talking 
		   about an empty file with the given number of channels
		   and sample rate. So for now, all is OK. Perhaps here
		   we could set a dummy 'untitled' filename.
		*/
		filename="";
	} else {
		filename=p_filename;
	}
	number_of_channels=channels;
}

ChiliFileDecoder::ChiliFileDecoder(ChiliSession * p_session) {
	log("ChiliFileDecoder::ChiliFileDecoder()");
	initvars();
	if (p_session!=NULL) {
		session=p_session;		
	}
	// dummy constructor for now.
}

ChiliFileDecoder::~ChiliFileDecoder() {
	log("ChiliFileDecoder::~ChiliFileDecoder()");
	
	// TODO: close all open soundfiles
	// TODO: clean up space allocated for (temporary) filenamesa
	if (data!=NULL) {
		free (data);
	}
	if (tmpdata!=NULL) {
		free (tmpdata);
	}
}

string ChiliFileDecoder::GetDecodedFilename() {
	log("ChiliFileDecoder::GetDecodedFilename()");
	return tmpfilename;
}
string ChiliFileDecoder::GetDecodedFilekey() {
	log("ChiliFileDecoder::GetDecodedFilekey()");
	return filekey;
}

unsigned int ChiliFileDecoder::GetNumberOfChannels() {
	log("ChiliFileDecoder::GetNumberOfChannels()");
	return number_of_channels;
}

void ChiliFileDecoder::LoadSourceFile(string p_filename) {
	// File supposedly contains info about sample rate and
	// amount of channels. This subroutine will set the properties
	// of the file object accordingly.
	log("ChiliFileDecoder::LoadSourceFile(%s)",p_filename);
	if (p_filename=="") {
		throw ERROR_FILENAME_OBLIGATORY;
	}
	filename=p_filename;

	SNDFILE         *infile;
	SF_INFO	sfinfo;

	if (! (infile = sf_open (filename.c_str(), SFM_READ, &sfinfo))) {
		log("Not able to open input file %s.\n", filename);
		throw ERROR_OPENING_INPUTFILE;
	}
	number_of_channels=sfinfo.channels;
	log ("number of channels in this file: %s",numtostr(number_of_channels).c_str());
	try {
        	decode_file(infile);
	}
	catch (int ERROR) {
		log("Decoding files failed.");
	}
	sf_close (infile) ;

	return;
}

